/*
 GBxMAS RW - LED controller
 PCB version: 1.0
 Firmware version: R1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 15/12/2019
 Last Modified: 5/01/2020
 
 Set fuse bits: External 16MHz crystal, divide clock by 8 is off
 avrdude -p atmega8515 -c usbasp -U lfuse:w:0xaf:m -U hfuse:w:0xda:m
 
 */

#define F_CPU 16000000 // 16 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include "setup.c" // See defines, variables, constants, functions here

volatile uint8_t setBlinkingProgress = 0;
volatile uint8_t currentBlinkingLED = 1;
volatile uint8_t blinkState = 1;
volatile uint8_t timerCounter = 0;
volatile uint16_t idleTimerCounter = 0;
volatile uint8_t idleDisplay = 0;
volatile uint8_t ledCounter = 0;
volatile uint8_t ledTop = 14;
volatile uint8_t ledBottom = 1;

int main(void) {
	setup();
	
	uint8_t randomLED = 0;
	uint8_t ledBuffer[18];

	// Idle timer on
	TIMSK1 |= (1<<TOIE1);
	TCNT1 = 0;
	TCCR1B |= (1<<CS12);
	
	// Timer for PWM breathing & blinking LEDs on
	TIMSK0 |= (1<<TOIE0);
	TCNT0 = 0;
	TCCR0A = (1<<CS02);
	uint8_t ledDirection = 1;
	uint8_t pwmCompare = 0;
	uint8_t timerUnset = 1;
	uint8_t redPWM = 0;
	
	while(1) {
		if (isIdle == 1) {
			sbi(UCSRB, RXCIE); // Enable USART RX interrupt
			
			if (idleDisplay == 0) { // Twinkling
				uint8_t ledAlreadySet = 1;
				while (ledAlreadySet == 1) {
					randomLED = rand() % 28;
					randomLED++;
					
					// Check LED isn't already set
					ledAlreadySet = 0;
					for (uint8_t x = 0; x < 18; x++) {
						if (randomLED == ledBuffer[x]) {
							ledAlreadySet = 1;
						}
					}
				}
				
				unsetLED(ledBuffer[ledCounter]);
				setLED(randomLED);
				ledBuffer[ledCounter] = randomLED;
				
				ledCounter++;
				if (ledCounter >= 18) {
					ledCounter = 0;
				}
				_delay_ms(100);
			}
			else if (idleDisplay == 1) { // Bottom left to top, top right to bottom
				for (uint8_t x = 1; x <= 14; x++) {
					setLED(x);
					_delay_ms(100);
					unsetLED(x);
				}
				for (uint8_t x = 28; x > 14; x--) {
					setLED(x);
					_delay_ms(100);
					unsetLED(x);
				}
				for (uint8_t x = 15; x <= 28; x++) {
					setLED(x);
					_delay_ms(100);
					unsetLED(x);
				}
				for (int8_t x = 14; x > 0; x--) {
					setLED(x);
					_delay_ms(100);
					unsetLED(x);
				}
			}
			else if (idleDisplay == 2) { // Tree left/right
				#define TREE_DELAY 50
				
				setLED(1);
				setLED(4);
				setLED(7);
				setLED(11);
				_delay_ms(TREE_DELAY);
				unsetLED(1);
				unsetLED(4);
				unsetLED(7);
				unsetLED(11);
				
				setLED(2);
				setLED(3);
				setLED(5);
				setLED(9);
				setLED(13);
				_delay_ms(TREE_DELAY);
				unsetLED(2);
				unsetLED(3);
				unsetLED(5);
				unsetLED(9);
				unsetLED(13);
				
				setLED(8);
				setLED(10);
				_delay_ms(TREE_DELAY);
				unsetLED(8);
				unsetLED(10);
				
				setLED(6);
				setLED(12);
				setLED(14);
				_delay_ms(TREE_DELAY);
				unsetLED(6);
				unsetLED(12);
				unsetLED(14);
				
				setLED(20);
				setLED(26);
				setLED(28);
				_delay_ms(TREE_DELAY);
				unsetLED(20);
				unsetLED(26);
				unsetLED(28);
				
				setLED(24);
				setLED(22);
				_delay_ms(TREE_DELAY);
				unsetLED(24);
				unsetLED(22);
				
				setLED(16);
				setLED(17);
				setLED(19);
				setLED(23);
				setLED(27);
				_delay_ms(TREE_DELAY);
				unsetLED(16);
				unsetLED(17);
				unsetLED(19);
				unsetLED(23);
				unsetLED(27);
				
				setLED(15);
				setLED(18);
				setLED(21);
				setLED(25);
				_delay_ms(TREE_DELAY);
				unsetLED(15);
				unsetLED(18);
				unsetLED(21);
				unsetLED(25);
			}
			else if (idleDisplay == 3) {
				for (int8_t x = ledTop; x > ledBottom; x--) {
					setLED(x);
					_delay_ms(100);
					unsetLED(x);
				}
				setLED(ledBottom);
				ledBottom++;
				
				if (ledBottom == 15) {
					ledTop = 28;
				}
				else if (ledBottom == 29) {
					ledTop = 14;
					ledBottom = 1;
					_delay_ms(2000);
					
					for (uint8_t x = 1; x <= 28; x++) {
						unsetLED(x);
					}
				}
			}
			else if (idleDisplay == 4) { // Breathing LEDs
				if (TCNT0 == 0) {
					PORTG &= ~(1<<LEFT_LED1);
					PORTB &= ~(1<<LEFT_LED3);
					PORTB &= ~(1<<LEFT_LED5);
					PORTB &= ~(1<<LEFT_LED7);
					PORTB &= ~(1<<LEFT_LED9);
					PORTF &= ~(1<<LEFT_LED12);
					PORTE &= ~(1<<LEFT_LED13);
					PORTC &= ~(1<<RIGHT_LED1);
					PORTC &= ~(1<<RIGHT_LED3);
					PORTC &= ~(1<<RIGHT_LED5);
					PORTA &= ~(1<<RIGHT_LED7);
					PORTA &= ~(1<<RIGHT_LED9);
					PORTA &= ~(1<<RIGHT_LED12);
					PORTA &= ~(1<<RIGHT_LED13);
					
					PORTB &= ~(1<<LEFT_LED2);
					PORTB &= ~(1<<LEFT_LED4);
					PORTG &= ~(1<<LEFT_LED6);
					PORTB &= ~(1<<LEFT_LED8);
					PORTB &= ~(1<<LEFT_LED10);
					PORTE &= ~(1<<LEFT_LED11);
					PORTF &= ~(1<<LEFT_LED14);
					PORTC &= ~(1<<RIGHT_LED2);
					PORTC &= ~(1<<RIGHT_LED4);
					PORTC &= ~(1<<RIGHT_LED6);
					PORTA &= ~(1<<RIGHT_LED8);
					PORTA &= ~(1<<RIGHT_LED10);
					PORTA &= ~(1<<RIGHT_LED11);
					PORTA &= ~(1<<RIGHT_LED14);
					timerUnset = 1;
				}
				
				if (TCNT0 == pwmCompare && timerUnset == 1) {
					if (redPWM == 1) {
						PORTG |= (1<<LEFT_LED1);
						PORTB |= (1<<LEFT_LED3);
						PORTB |= (1<<LEFT_LED5);
						PORTB |= (1<<LEFT_LED7);
						PORTB |= (1<<LEFT_LED9);
						PORTF |= (1<<LEFT_LED12);
						PORTE |= (1<<LEFT_LED13);
						PORTC |= (1<<RIGHT_LED1);
						PORTC |= (1<<RIGHT_LED3);
						PORTC |= (1<<RIGHT_LED5);
						PORTA |= (1<<RIGHT_LED7);
						PORTA |= (1<<RIGHT_LED9);
						PORTA |= (1<<RIGHT_LED12);
						PORTA |= (1<<RIGHT_LED13);
					}
					else {
						PORTB |= (1<<LEFT_LED2);
						PORTB |= (1<<LEFT_LED4);
						PORTG |= (1<<LEFT_LED6);
						PORTB |= (1<<LEFT_LED8);
						PORTB |= (1<<LEFT_LED10);
						PORTE |= (1<<LEFT_LED11);
						PORTF |= (1<<LEFT_LED14);
						PORTC |= (1<<RIGHT_LED2);
						PORTC |= (1<<RIGHT_LED4);
						PORTC |= (1<<RIGHT_LED6);
						PORTA |= (1<<RIGHT_LED8);
						PORTA |= (1<<RIGHT_LED10);
						PORTA |= (1<<RIGHT_LED11);
						PORTA |= (1<<RIGHT_LED14);
					}
					
					if (ledDirection == 1) {
						pwmCompare++;
						if (pwmCompare >= 254) {
							ledDirection = 0;
							redPWM = !redPWM;
						}
					}
					else if (ledDirection == 0) {
						pwmCompare--;
						if (pwmCompare == 0) {
							ledDirection = 1;
						}
					}
					timerUnset = 0;
				}
			}
			
			/*else if (idleDisplay == 4) { // Blinking eyes (not included by default)
				setLED(12);
				setLED(26);
				_delay_ms(2000);
				unsetLED(12);
				unsetLED(26);
				_delay_ms(250);
				//setLED(9);
				//setLED(23);
				setLED(6);
				setLED(8);
				setLED(22);
				setLED(20);
			}*/
			else if (idleDisplay == 10) { // Circle, used for chip erase (not an idle animation)
				ledBuffer[0] = 12;
				ledBuffer[1] = 26;
				ledBuffer[2] = 24;
				ledBuffer[3] = 23;
				ledBuffer[4] = 22;
				ledBuffer[5] = 20;
				ledBuffer[6] = 6;
				ledBuffer[7] = 8;
				ledBuffer[8] = 9;
				ledBuffer[9] = 10;
				
				for (uint8_t x = 0; x < 10; x++) {
					setLED(ledBuffer[x]);
					_delay_ms(100);
					unsetLED(ledBuffer[x]);
				}
			}
		}
		
		// Packet received or we performed a WDT reset
		else {
			allLEDOff();
			
			while (isIdle == 0) {
				receivedChar = USART_Receive(); 
				
				if (receivedChar == XMAS_LEDS) {
					usart_read_chars();
					uint32_t xmasValueCheck = strtol(receivedBuffer, NULL, 16); 
					
					// Verified ok
					if (xmasValueCheck == XMAS_VALUE) {
						receivedChar = USART_Receive(); // Wait for next command
						
						if (receivedChar == 'L') { // Set/unset LEDs
							usart_read_chars(); // Read 32bit number
							uint32_t ledBits = strtol(receivedBuffer, NULL, 16); // Convert address string in hex to dec
							
							// Check which bits are set
							for (uint8_t b = 0; b < 28; b++) {
								uint8_t ledNum = b + 1;
								if ((uint32_t) ledBits & (uint32_t) ((uint32_t) 1 << (uint32_t) b)) {
									setLED(ledNum);
								}
								else {
									unsetLED(ledNum);
								}
							}
							
							setBlinkingProgress = 0;
						}
						else if (receivedChar == 'B') {
							currentBlinkingLED = USART_Receive();
							setBlinkingProgress = 1;
							
							// Timer on
							TIMSK0 |= (1<<TOIE0);
							TCNT0 = 0;
							TCCR0A = (1<<CS02);
						}
						else if (receivedChar == 'I') { // Turn on idle counter
							TIMSK1 |= (1<<TOIE1);
							TCNT1 = 0;
							TCCR1B |= (1<<CS12);
						}
						else if (receivedChar == 'O') { // Turn off idle counter
							TCCR1B &= ~(1<<CS12);
						}
						else if (receivedChar == 'E') { // Force circle animation for chip erase
							TCCR1B &= ~(1<<CS12);
							isIdle = 1;
							idleDisplay = 10;
						}
					}
					idleTimerCounter = 0; // Reset idle timer
				}
			}
			allLEDOff();
			randomLED = 0;
			ledCounter = 0;
			ledTop = 14;
			ledBottom = 1;
		}
	}
}

ISR(TIMER0_OVF_vect) {
	timerCounter++;
	if (timerCounter >= 20) {
		if (setBlinkingProgress == 1) {
			if (blinkState == 1) {
				setLED(currentBlinkingLED);
				blinkState = 0;
			}
			else {
				unsetLED(currentBlinkingLED);
				blinkState = 1;
			}
		}
		timerCounter = 0;
	}
}

ISR(TIMER1_OVF_vect) {
	idleTimerCounter++;
	if (idleTimerCounter >= 110) { // ~2 mins
		idleDisplay++;
		ledCounter = 0;
		ledTop = 14;
		ledBottom = 1;
		allLEDOff();
		
		if (idleDisplay > 4) {
			idleDisplay = 0;
		}
		idleTimerCounter = 0;
		isIdle = 1;
	}
}

ISR(USART0_RX_vect) {
	if (UDR == '!') {
		// Clear watchdog flag
		MCUSR &= ~(1<<WDRF);
		// Start timed sequence
		WDTCR = (1<<WDCE) | (1<<WDE);
		// Reset in 30 ms
		WDTCR = (1<<WDP0) | (1<<WDE);
		// Wait for reset
		_delay_loop_2(65535);
	}
}